/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2024  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/times.h>
#include <pwd.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "xmem.h"
#include "config.h"

// ****************************************************************************
// 
// ****************************************************************************
ben_config conf;

// ****************************************************************************
// 
// ****************************************************************************
static char *xrandom(size_t len)
{
	char *s;
	size_t i;
	int v;
	
	s = xmalloc(len + 1);
	
	for (i = 0; i < len; i++) {
		v = rand() % (10 + 26 + 26);
		
		if (v < 10) {
			v = '0' + v;
		} else if (v < 10 + 26) {
			v = 'A' + (v - 10);
		} else {
			v = 'a' + (v - (10 + 26));
		}
		
		s[i] = v;
	}
	
	s[len] = 0;
	return(s);
}

// ****************************************************************************
// 
// ****************************************************************************
void config_init(void)
{
	static char buf[1025];
	struct passwd *pwd;
	char *user;
	char *s;
	long int pid, uid;
	
	conf.file_size_max = BEN_DEFAULT_FILE_SIZE_MAX;
	conf.proto_buff_max = BEN_DEFAULT_PROTO_BUFF_MAX;
	conf.block_size = BEN_DEFAULT_BLOCK_SIZE;
	conf.default_port = BEN_DEFAULT_PORT;
	
	conf.cpus_online = sysconf(_SC_NPROCESSORS_ONLN);
	if ((conf.cpus_online <= 0) || (conf.cpus_online > 65536))
		conf.cpus_online = 1;
	
	conf.keepalive_ms = BEN_DEFAULT_KEEPALIVE_MS;
	conf.keepalive_slack = BEN_DEFAULT_KEEPALIVE_SLACK;
	conf.snapshot_ms = BEN_DEFAULT_SNAPSHOT_MS;
	
	s = getenv("SHELL");
	if (!s)
		s = BEN_DEFAULT_SHELL;
	conf.shell = xstrdup(s);
	
	s = getenv("HOSTNAME");
	if (s) {
		conf.host = xstrdup(s);
	} else {
		if (gethostname(buf, sizeof(buf))) {
			fprintf(stderr, "gethostname(): %s\n", strerror(errno));
			errno = 0;
			conf.host = xstrdup(BEN_DEFAULT_HOST);
		} else {
			buf[sizeof(buf) - 1] = 0;
			conf.host = xstrdup(buf);
		}
	}
	
	uid = (long int)getuid();
	pid = (long int)getpid();
	
	pwd = getpwuid(uid);
	
	if (pwd) {
		user = pwd->pw_name;
	} else {
		fprintf(stderr, "getpwuid(%ld): %s\n", uid, strerror(errno));
		errno = 0;
		
		user = getenv("USER");
		
		if (!user)
			user = "user";
	}
	
	conf.user = xstrdup(user);
	
	srand(times(NULL));
	conf.random_id = xrandom(6);
	
	conf.server_log = xprintf(NULL,
		"/tmp/ben-%s/server-%ld.log", conf.user, pid);
	conf.client_log = xprintf(NULL,
		"/tmp/ben-%s/client-%ld.log", conf.user, pid);
	conf.remote_log = xprintf(NULL,
		"/tmp/ben-%s/remote-%s.log", conf.user, conf.random_id);
	
	if (!isatty(1)) {
		errno = 0;
		conf.color = BEN_COLOR_NONE;
	} else if (getenv("NO_COLOR")) {
		conf.color = BEN_COLOR_NONE;
	} else {
		conf.color = BEN_COLOR_TTY;
	}
}

// ****************************************************************************
// 
// ****************************************************************************
void config_clear(void)
{
	xfree(conf.shell);
	xfree(conf.host);
	xfree(conf.user);
	xfree(conf.random_id);
	xfree(conf.server_log);
	xfree(conf.client_log);
	xfree(conf.remote_log);
}


// ****************************************************************************
// 
// ****************************************************************************

