/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2024  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "xmem.h"
#include "date.h"

// ****************************************************************************
// 
// ****************************************************************************
gtime gtime_zero(void)
{
	gtime r;
	
	r.tv_sec = 0;
	r.tv_nsec = 0;
	
	return r;
}

gtime gtime_wall(void)
{
	gtime r;
	
	if (clock_gettime(CLOCK_REALTIME, &r)) {
		fprintf(stderr, "clock_getttime(CLOCK_REALTIME): %s\n",
			strerror(errno));
		exit(1);
	}

	return r;
}

gtime gtime_mono(void)
{
	gtime r;
	
	if (clock_gettime(CLOCK_MONOTONIC, &r)) {
		fprintf(stderr, "clock_getttime(CLOCK_MONOTONIC): %s\n",
			strerror(errno));
		exit(1);
	}
	
	return r;
}

gtime gtime_s(uint64_t s)
{
	gtime r;
	
	r.tv_sec = s;
	r.tv_nsec = 0;
	
	return r;
}

gtime gtime_ms(uint64_t ms)
{
	gtime r;
	
	r.tv_sec = ms / 1000;
	r.tv_nsec = (uint32_t)(ms % 1000) * 1000000;
	
	return r;
}

gtime gtime_add(gtime a, gtime b)
{
	gtime r;
	
	r.tv_sec = a.tv_sec + b.tv_sec;
	r.tv_nsec = a.tv_nsec + b.tv_nsec;
	
	if (r.tv_nsec > 1000000000) {
		r.tv_nsec -= 1000000000;
		r.tv_sec++;
	}
	
	return r;
}

gtime gtime_sub(gtime a, gtime b)
{
	gtime r;
	
	if (a.tv_nsec > b.tv_nsec) {
		r.tv_nsec = a.tv_nsec - b.tv_nsec;
		r.tv_sec = a.tv_sec - b.tv_sec;
	} else {
		r.tv_nsec = 1000000000 + a.tv_nsec - b.tv_nsec;
		r.tv_sec = a.tv_sec - b.tv_sec - 1;
	}
	
	return r;
}

int gtime_nonzero(gtime a)
{
	return (a.tv_nsec || a.tv_sec);
}

int gtime_ge(gtime a, gtime b)
{
	if (a.tv_sec > b.tv_sec)
		return 1;
	
	if (a.tv_sec < b.tv_sec)
		return 0;
	
	return (a.tv_nsec >= b.tv_nsec);
}

uint64_t gtime_get_s(gtime a)
{
	return a.tv_sec;
}

uint64_t gtime_get_ms(gtime a)
{
	return a.tv_sec * 1000 + (a.tv_nsec / 1000000);
}

uint64_t gtime_get_ns(gtime a)
{
	return a.tv_sec * 1000000000 + a.tv_nsec;
}

// ****************************************************************************
// 
// ****************************************************************************
gtime_str duration_str(gtime dur, int flags)
{
	gtime_str r;
	time_t t = dur.tv_sec;
	
	int ds = t % 60;
	t /= 60;
	
	int dm = t % 60;
	t /= 60;
	
	size_t offs;
	int dh, dd;

	if (flags & GTIME_DAYS) {
		dh = t % 24;
		dd = t / 24;

		if (dd) {
			snprintf(r.s, sizeof(r.s),
				"%4dd %02d:%02d:%02d", dd, dh, dm, ds);
		} else {
			snprintf(r.s, sizeof(r.s),
				"      %02d:%02d:%02d", dh, dm, ds);
		}
		
		offs = 13;
	} else {
		dh = t;

		snprintf(r.s, sizeof(r.s), "%02d:%02d:%02d", dh, dm, ds);
		
		offs = 8;
	}
	
	if (flags & GTIME_SUBSEC) {
		snprintf(r.s + offs, sizeof(r.s) - offs,
			".%09lu", (unsigned long)dur.tv_nsec);
	}
	
	r.s[sizeof(r.s) - 1] = 0;
	
	return r;
}

gtime_str date_str(gtime t, int flags)
{
	gtime_str r;
	time_t ts = t.tv_sec;
	
	struct tm bd;
	struct tm *ptr = localtime_r(&ts, &bd);
	
	if (!ptr) {
		fprintf(stderr, "localtime_r(): %s\n", strerror(errno));
		exit(1);
	}
	
	snprintf(r.s, sizeof(r.s),
		"%04d-%02d-%02d %02d:%02d:%02d",
		1900 + bd.tm_year, 1 + bd.tm_mon, bd.tm_mday,
		bd.tm_hour, bd.tm_min, bd.tm_sec);
	
	if (flags & GTIME_SUBSEC) {
		snprintf(r.s + 19, sizeof(r.s) - 19,
			".%09lu", (unsigned long)t.tv_nsec);
	}
	
	r.s[sizeof(r.s) - 1] = 0;
	
	return r;
}

gtime_str curdate_str(int flags)
{
	return date_str(gtime_wall(), flags);
}


// ****************************************************************************
// 
// ****************************************************************************
