/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2024  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef BEN_BOX_H
#define BEN_BOX_H
#include <stdint.h>
#include "date.h"
#include "buffer.h"
#include "network.h"
#include "queue.h"

/*
node flags:

  NODE_KICK:
    kick this node

  NODE_RETIRE:
    set NODE_KICK once running.size == 0

  NODE_SYNC_PENDING:
    earlier pending sync job on this node
    - cleared at the beginning of every scheduling cycle
    - set when a sync job is seen

  NODE_SYNC_RUNNING:
    running sync job on this node
    - cleared when running.size == 0
    - set when the sync job is run

  NODE_SYNC_GLOBAL:
    running global sync job on this or another node
    - cleared when running.size == 0 and NODE_SYNC_RUNNING
    - set when a global sync job is run

  NODE_SYNC_ANY:
    SYNC_ PENDING | RUNNING | GLOBAL
*/

#define NODE_KICK		1
#define NODE_RETIRE		2
#define NODE_SYNC_PENDING	4
#define NODE_SYNC_RUNNING	8
#define NODE_SYNC_GLOBAL	16
#define NODE_SYNC_ANY		\
		(NODE_SYNC_PENDING | NODE_SYNC_RUNNING | NODE_SYNC_GLOBAL)

typedef struct node_t {
	int fd, id;
	char *name;
	int processes, flags;
	int refs;
	
	gtime atime;
	uint64_t ping_sent, ping_received;
	
	int ack_h;
	size_t ack_n;
	
	buffer inbuff;
	buffer outbuff;
	
	queue running;

	struct node_t *prev;
	struct node_t *next;
} node;

node *node_new(int id, int fd);
void node_free(node *b);

void node_setup(node *b, char *name, int cores);

int node_matches(node *b, pattern *pat);

typedef struct nodeset_t {
	node *head;
	node *tail;
	int n;
} nodeset;

void nodeset_init(nodeset *bs);
void nodeset_clear(nodeset *bs);

void nodeset_attach(nodeset *bs, node *b);
void nodeset_detach(nodeset *bs, node *b);

node *nodeset_find(nodeset *bs, int id);

#endif

