/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2024  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>
#include "xmem.h"
#include "block.h"
#include "date.h"
#include "array.h"
#include "buffer.h"
#include "pattern.h"
#include "config.h"
#include "network.h"
#include "queue.h"
#include "node.h"
#include "common.h"
#include "protocol.h"
#include "exec.h"
#include "snapshot.h"
#include "server.h"

// ***************************************************************************
// 
// ***************************************************************************
#define SERVER_TARGET_RUNNING	0
#define SERVER_TARGET_EXIT	1
#define SERVER_TARGET_RETIRE	2

typedef struct server_t {
	int fd;
	nodeset clients;
	queue done, pending;
	array pfd_a;
	gtime wall_time_cur;
	gtime mono_time_cur;
	gtime mono_time_snap;
	
	int box_pool, task_pool;
	int target;
	int modified;
	
	char *snapshot_path;
} server;

#define DEBUG(...)	\
	do { \
		fprintf(stderr, "%s ", \
			curdate_str(GTIME_DAYS | GTIME_SUBSEC).s); \
		fprintf(stderr, __VA_ARGS__); \
	} while(0)

// ***************************************************************************
// 
// ***************************************************************************
static void task_describe(buffer *buff,
	server *g, node *c, job *t, int type)
{
	node *fc;
	int bid;
	char *bname;
	uint64_t t0, t1;
	
	if (c) {
		bid = c->id;
		bname = c->name;
	} else if ((t->restrict_box >= 0) || (t->ran_box >= 0)) {
		bid = (t->restrict_box >= 0) ? t->restrict_box : t->ran_box;

		fc = nodeset_find(&g->clients, bid);
		if (fc)
			bname = fc->name;
		else
			bname = "-";
	} else {
		bid = -1;
		bname = "";
	}
	
	t0 = gtime_get_s(t->wall_time_start);
	t1 = gtime_get_s((type == PROTO_TASK_RUNNING)
		? g->wall_time_cur
		: t->wall_time_stop);
	
	proto_append(buff, PROTO_SV_MAGIC, PROTO_SV_JOB,
		"issiiispp",
		t->id, t->dir, t->name, type,
		t->flags, bid, bname,
		t0, t1);
}

// ***************************************************************************
// 
// ***************************************************************************
static void serv_fwd_unref(server *g, job *t)
{
	node *fc;
	
	if (t->forward_box < 0)
		return;
		
	fc = nodeset_find(&g->clients, t->forward_box);
	
	if (fc) {
		fc->refs--;
	} else {
		DEBUG("[%d %s %s]: no fwd client (%d), EOF dismissed\n",
			t->id, t->dir, t->name, t->forward_box);
	}
}

// ***************************************************************************
// 
// ***************************************************************************
static int serv_sched_client(server *g, node *c)
{
	node *fc;
	job *t, *tnext;
	void *d;
	size_t dsz;
	char msg[128];
	char exec_str[32];
	pattern pat_dir, pat_job, pat_node;
	char *s, *hs, *shell, *dir, *opts, *name, *cmd, *bname;
	int h, cores, id, flags, fd, rcode, count, found;
	
	if (c->outbuff.size > conf.proto_buff_max) {
		DEBUG("(%d %s): output buffer too large (%zd > %zd)\n",
			c->id, c->name, c->outbuff.size, conf.proto_buff_max);
		return(-1);
	}

	if (proto_prepare(&c->inbuff, PROTO_CL_MAGIC, conf.proto_buff_max)) {
		DEBUG("(%d %s): protocol error\n", c->id, c->name);
		return(-1);
	}
	
	if (!proto_ready(&c->inbuff)) {
		gtime tx = gtime_add(c->atime, gtime_ms(conf.keepalive_ms));
		
		if (gtime_ge(g->mono_time_cur, tx)) {
			if (c->ping_received != c->ping_sent) {
				DEBUG("(%d %s): not responding "
					"(sent %ld, received %ld)\n",
					c->id, c->name,
					(long)c->ping_sent,
					(long)c->ping_received);
			}
			
			if ((c->ping_received > c->ping_sent)
			||  (c->ping_received + conf.keepalive_slack
					< c->ping_sent)) {
				DEBUG("(%d %s): removed\n", c->id, c->name);
				return(-1);
			}
			
			c->ping_sent++;
			proto_append(&c->outbuff, PROTO_SV_MAGIC,
				PROTO_SV_PING, "p", c->ping_sent);
			c->atime = g->mono_time_cur;

			//DEBUG("(%d %s): sending ping %zd\n",
			//	c->id, c->host, c->ping_sent);
		}
		return(0);
	}
	
	h = proto_header(&c->inbuff);

	// **********************************
	if (h == PROTO_CL_SETUP) {
		if (proto_fetch(&c->inbuff, "is", &cores, &s)) {
			DEBUG("(%d %s): PROTO_CL_SETUP: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		DEBUG("(%d %s): setup as (%d %s) with %d processes\n",
			c->id, c->name, c->id, s, cores);
		
		node_setup(c, s, cores);
		return(0);
	}
		
	// **********************************
	if ((h == PROTO_CL_OUT) || (h == PROTO_CL_ERR)) {
		hs = (h == PROTO_CL_OUT) ? "OUT" : "ERR";
		
		if (proto_fetch(&c->inbuff, "id", &id, &d, &dsz)) {
			DEBUG("(%d %s): PROTO_CL_%s: bad packet\n",
				c->id, c->name, hs);
			return(-1);
		}

		t = queue_find(&c->running, id);
		if (!t) {
			DEBUG("(%d %s): PROTO_CL_%s: no task id %d\n",
				c->id, c->name, hs, id);
			return(-1);
		}
		
		if (h == PROTO_CL_OUT)
			t->total_out += dsz;
		else
			t->total_err += dsz;
		
		// Forward data for task t from c to fc
		if (t->forward_box >= 0) {
			fc = nodeset_find(&g->clients, t->forward_box);
			
			if (fc) {
				proto_append(&fc->outbuff, PROTO_SV_MAGIC,
					(h == PROTO_CL_OUT) ? PROTO_SV_OUT
						: PROTO_SV_ERR,
					"iisd", t->id, c->id, c->name, d, dsz);
			} else {
				DEBUG("[%d %s %s]: no fwd client (%d), "
					"data dismissed\n",
					t->id, t->dir, t->name, t->forward_box);
			}
		}
		
		// Write data for task t on server-side storage
		fd = (h == PROTO_CL_OUT) ? t->outfd : t->errfd;
		
		if (fd >= 0) {
			if (block_write(fd, d, dsz)) {
				fprintf(stderr, "%s: %s: %s\n", t->name,
					(h == PROTO_CL_OUT) ? t->outpath
						: t->errpath,
					strerror(errno));
				exit(1);
			}
		}
		
		return(0);
	}
	
	// **********************************
	if (h == PROTO_CL_END) {
		if (proto_fetch(&c->inbuff, "ii", &id, &rcode)) {
			DEBUG("(%d %s): PROTO_CL_END: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		t = queue_find(&c->running, id);
		if (!t) {
			DEBUG("(%d %s): PROTO_CL_END: no task [%d]\n",
				c->id, c->name, id);
			return(-1);
		}

		DEBUG("[%d %s %s]: finished on (%d %s): %s; "
			"totals %zd %zd\n",
			t->id, t->dir, t->name, c->id, c->name,
			(rcode == PROTO_END_DONE) ? "done"
			: ((rcode == PROTO_END_STOP) ? "stop" : "fail"),
			t->total_out, t->total_err);
			
		// Close server-local files:
		// if rcode is DONE, mark task as done
		// if rcode is STOP, delete task
		// if rcode is FAIL, remove files and requeue task

		if (rcode == PROTO_END_DONE) {
			job_close(t, 0);
			serv_fwd_unref(g, t);
			queue_unlink(&c->running, t);
			t->ran_box = c->id;
			queue_append(&g->done, t);
		} else if (rcode == PROTO_END_STOP) {
			job_close(t, 0);
			serv_fwd_unref(g, t);
			queue_unlink(&c->running, t);
			job_free(t);
		} else if (rcode == PROTO_END_FAIL) {
			job_close(t, 1);
			queue_unlink(&c->running, t);
			queue_prepend(&g->pending, t);
		}
		
		g->modified = 1;
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_PONG) {
		if (proto_fetch(&c->inbuff, "p", &c->ping_received)) {
			DEBUG("(%d %s): PROTO_CL_PONG: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		if ((c->ping_received > c->ping_sent)
		|| (c->ping_received + conf.keepalive_slack < c->ping_sent)) {
			DEBUG("(%d %s): PROTO_CL_PONG: wrong response "
				"(%ld != %ld)\n",
				c->id, c->name,
				(long)c->ping_received,
				(long)c->ping_sent);
			return(-1);
		}

		//DEBUG("(%d %s): ping response %zd\n",
		//	c->id, c->host, c->ping_received);
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_REQACK) {
		if (proto_fetch(&c->inbuff, "")) {
			DEBUG("(%d %s): PROTO_CL_REQACK: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		if (c->ack_h == PROTO_CL_LIST) {
			count = 0;
		} else {
			count = xsnprintf(msg, sizeof(msg), "%s: %zd\n",
				(c->ack_h == PROTO_CL_ADD) ? "Queued" :
				(c->ack_h == PROTO_CL_RM) ? "Removed" :
				"Matched", c->ack_n);
		}
		
		c->ack_h = -1;
		c->ack_n = 0;
		
		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg, (size_t)count);
		
		// - if c->refs == 1 (normal state after add, rm, scale, list),
		//   then c->refs becomes 0, which kicks client after last
		//   i/o byte has been written.
		// - if c->refs > 1 (after some resume with restrict_name),
		//   then client gets kicked after output from execs.
		c->refs--;

		return(0);
	}

	// **********************************
	if (h == PROTO_CL_ADD) {
		if (proto_fetch(&c->inbuff, "sssss", &shell,
				&dir, &opts, &name, &cmd)) {
			DEBUG("(%d %s): PROTO_CL_ADD: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}
		c->ack_n++;
		
		t = job_new(g->task_pool++, shell, dir, opts, name, cmd);
		queue_append(&g->pending, t);
		
		DEBUG("[%d %s %s]: added to queue\n",
			t->id, t->dir, t->name);

		g->modified = 1;
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_EXEC) {
		if (proto_fetch(&c->inbuff, "sss", &shell, &opts, &cmd)) {
			DEBUG("(%d %s): PROTO_CL_EXEC: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		c->refs = 0;
		
		// count clients with cores
		cores = 0;
		fc = g->clients.head;
		
		while (fc) {
			if (fc->processes >= 0)
				cores++;
			fc = fc->next;
		}

		// task name
		xsnprintf(exec_str, sizeof(exec_str), "exec-%d-%d",
			g->task_pool, g->task_pool + cores - 1);
		
		// enqueue
		fc = g->clients.head;
		
		while (fc) {
			if (fc->processes >= 0) {
				t = job_new(g->task_pool++, shell, "", opts,
					exec_str, cmd);
				t->flags = PROTO_JOB_SYNC;
				t->restrict_box = fc->id;
				t->forward_box = c->id;
				c->refs++;
			
				queue_append(&g->pending, t);
				
				DEBUG("[%d %s %s]: queued for sync "
					"on (%d %s)\n",
					t->id, t->dir, exec_str,
					fc->id, fc->name);
				g->modified = 1;
			}
			
			fc = fc->next;
		}

		
		count = xsnprintf(msg, sizeof(msg),
			"%s: Queued %d sync tasks (ctrl+c to discard output)\n",
			exec_str, cores);
		
		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg, (size_t)count);

		return(0);
	}

	// **********************************
	if (h == PROTO_CL_RM) {
		if (proto_fetch(&c->inbuff, "iss", &flags, &dir, &name)) {
			DEBUG("(%d %s): PROTO_CL_RM: bad packet\n",
				c->id, c->name);
			return(-1);
		}

		// no short-circuit "or", so that both get initialized
		if (pattern_init(&pat_dir, dir, 0)
		|   pattern_init(&pat_job, name, PATTERN_RANGES)) {
			pattern_clear(&pat_dir);
			pattern_clear(&pat_job);
			DEBUG("(%d %s): PROTO_CL_RM: bad pattern %s %s\n",
				c->id, c->name, dir, name);
			return(-1);
		}

		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}
		
		// For all matching done tasks, remove
		t = (flags & PROTO_TASK_DONE) ? g->done.head : NULL;
		while (t) {
			tnext = t->next;
			
			if (job_matches(t, &pat_dir, &pat_job)) {
				queue_unlink(&g->done, t);
				job_free(t);
				c->ack_n++;
				g->modified = 1;
			}
			
			t = tnext;
		}

		// For all matching running tasks, send stop cmd
		fc = (flags & PROTO_TASK_RUNNING) ? g->clients.head : NULL;
		while (fc) {
			t = fc->running.head;
			while (t) {
				if (job_matches(t, &pat_dir, &pat_job)) {
					proto_append(&fc->outbuff,
						PROTO_SV_MAGIC, PROTO_SV_STOP,
						"ii", t->id,
						PROTO_END_STOP);
					DEBUG("[%d %s %s]: interrupting"
						" on (%d %s)\n",
						t->id, t->dir, t->name,
						fc->id, fc->name);
					c->ack_n++;
				}
				t = t->next;
			}
			fc = fc->next;
		}
		
		// For all matching pending tasks, remove
		t = (flags & PROTO_TASK_PENDING) ? g->pending.head : NULL;
		while (t) {
			tnext = t->next;
		
			if (job_matches(t, &pat_dir, &pat_job)) {
				serv_fwd_unref(g, t);
				queue_unlink(&g->pending, t);
				job_free(t);
				c->ack_n++;
				g->modified = 1;
			}
		
			t = tnext;
		}
		
		pattern_clear(&pat_dir);
		pattern_clear(&pat_job);
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_RESUME) {
		if (proto_fetch(&c->inbuff, "ssssssi", &shell,
				&dir, &opts, &name, &cmd, &bname, &flags)) {
			DEBUG("(%d %s): PROTO_CL_RESUME: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		// PROTO_CL_RESUME shares ack_n with PROTO_CL_ADD
		if (c->ack_h != PROTO_CL_ADD) {
			c->ack_h = PROTO_CL_ADD;
			c->ack_n = 0;
		}
		
		// find target client
		id = -1;
		
		if (bname[0] != 0) {
			fc = g->clients.head;
			found = 0;
			
			while (fc) {
				if (!strcmp(fc->name, bname)) {
					if (id < 0)
						id = fc->id;
					found++;
				}
				
				fc = fc->next;
			}
			
			if (found != 1) {
				DEBUG("[x %s %s]: client '%s': %d found\n",
					dir, name, bname, found);
				
				count = xsnprintf(msg, sizeof(msg),
					"%s %s: client '%s': %d found\n",
					dir, name, bname, found);
				
				proto_append(&c->outbuff, PROTO_SV_MAGIC,
					PROTO_SV_MSG, "d", msg, (size_t)count);
				
				return 0;
			}
		}
		
		// enqueue
		c->ack_n++;
		
		t = job_new(g->task_pool++, shell, dir, opts, name, cmd);
		
		t->flags = flags;
		t->restrict_box = id;
		
		if (t->flags & PROTO_JOB_SYNC) {
			t->forward_box = c->id;
			c->refs++;
			
			count = xsnprintf(msg, sizeof(msg),
				"%s: Queued sync task on '%s'"
				" (ctrl+c to discard output)\n",
				name, bname);
			
			proto_append(&c->outbuff, PROTO_SV_MAGIC,
				PROTO_SV_MSG, "d", msg, (size_t)count);
		}
		
		queue_append(&g->pending, t);
		
		DEBUG("[%d %s %s]: resumed to queue on (%d %s)\n",
			t->id, t->dir, t->name,
			id, (id >= 0) ? bname : "<any>");

		g->modified = 1;
		return(0);
	}
	
	// **********************************
	if (h == PROTO_CL_SCALE) {
		if (proto_fetch(&c->inbuff, "sii", &name, &cores, &flags)) {
			DEBUG("(%d %s): PROTO_CL_SCALE: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		if (cores < 0)
			cores = 0;
		
		if (pattern_init(&pat_node, name, PATTERN_RANGES)) {
			DEBUG("(%d %s): PROTO_CL_SCALE: bad pattern %s\n",
				c->id, c->name, name);
			return(-1);
		}

		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}
		
		fc = g->clients.head;
		while (fc) {
			// no scaling of control clients (kill allowed)
			if ((fc->processes < 0)
			&&  (!(flags & PROTO_SCALE_KILL))) {
				fc = fc->next;
				continue;
			}
			
			// no suicide allowed, match id, match name
			if ((fc->id == c->id)
			||  (!node_matches(fc, &pat_node))) {
				fc = fc->next;
				continue;
			}

			c->ack_n++;
			
			if (flags & PROTO_SCALE_KILL)
				fc->flags |= NODE_RETIRE;

			count = fc->running.size - cores;
			
			if (!(flags & PROTO_SCALE_RETIRE)) {
				t = fc->running.tail;
				while (count > 0) {
					proto_append(&fc->outbuff,
						PROTO_SV_MAGIC, PROTO_SV_STOP,
						"ii", t->id,
						PROTO_END_FAIL);
					DEBUG("[%d %s %s]: interrupting"
						" on (%d %s): over capacity "
						"after scaling\n",
						t->id, t->dir, t->name,
						fc->id, fc->name);
					count--;
					t = t->prev;
				}
			}
			
			if (fc->processes >= 0)
				fc->processes = cores;
			
			fc = fc->next;
		}
		
		pattern_clear(&pat_node);

		return(0);
	}

	// **********************************
	if (h == PROTO_CL_LIST) {
		if (proto_fetch(&c->inbuff, "iss", &flags, &dir, &name)) {
			DEBUG("(%d %s): PROTO_CL_LIST: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		// no short-circuit "or", so that both get initialized
		if (pattern_init(&pat_dir, dir, 0)
		|   pattern_init(&pat_job, name, PATTERN_RANGES)) {
			pattern_clear(&pat_dir);
			pattern_clear(&pat_job);
			DEBUG("(%d %s): PROTO_CL_LIST: bad pattern %s %s\n",
				c->id, c->name, dir, name);
			return(-1);
		}
		
		if (c->ack_h != h) {
			c->ack_h = h;
			c->ack_n = 0;
		}
		
		if (flags & PROTO_TASK_DONE) {
			t = g->done.head;
			while (t) {
				if (job_matches(t, &pat_dir, &pat_job)) {
					task_describe(&c->outbuff,
						g, NULL, t, PROTO_TASK_DONE);
					c->ack_n++;
				}
				
				t = t->next;
			}
		}

		if (flags & PROTO_TASK_RUNNING) {
			fc = g->clients.head;
			while (fc) {
				t = fc->running.head;
				while (t) {
					if (job_matches(t, &pat_dir,
							&pat_job)) {
						task_describe(&c->outbuff,
							g, fc, t,
							PROTO_TASK_RUNNING);
						c->ack_n++;
					}
					
					t = t->next;
				}
				
				fc = fc->next;
			}
		}

		if (flags & PROTO_TASK_PENDING) {
			t = g->pending.head;
			while (t) {
				if (job_matches(t, &pat_dir, &pat_job)) {
					task_describe(&c->outbuff,
						g, NULL, t,
						PROTO_TASK_PENDING);
					c->ack_n++;
				}
				
				t = t->next;
			}
		}
		
		pattern_clear(&pat_dir);
		pattern_clear(&pat_job);
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_NODES) {
		if (proto_fetch(&c->inbuff, "")) {
			DEBUG("(%d %s): PROTO_CL_NODES: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		fc = g->clients.head;
		while (fc) {
			proto_append(&c->outbuff,
				PROTO_SV_MAGIC, PROTO_SV_NODE,
				"isiippii",
				fc->id, fc->name,
				fc->running.size, fc->processes,
				fc->inbuff.size, fc->outbuff.size,
				fc->refs, fc->flags);
			fc = fc->next;
		}

		c->refs = 0;
		return(0);
	}

	// **********************************
	if (h == PROTO_CL_EXIT) {
		if (proto_fetch(&c->inbuff, "i", &flags)) {
			DEBUG("(%d %s): PROTO_CL_EXIT: bad packet\n",
				c->id, c->name);
			return(-1);
		}
		
		// The exit message will not display client-side,
		// except when retiring with some tasks running.
		count = xsnprintf(msg, sizeof(msg), "%s",
			(flags & PROTO_SCALE_RETIRE)
			? "Retiring...\n"
			: "Exit.\n");

		proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_MSG,
			"d", msg, (size_t)count);
		
		// Keep client referenced, so that it hangs until
		// the exit is completed.
		//c->refs = 1;
		
		if (flags & PROTO_SCALE_RETIRE)
			g->target |= SERVER_TARGET_RETIRE;
		else
			g->target |= SERVER_TARGET_EXIT;

		return(0);
	}

	// **********************************
	DEBUG("(%d %s): unknown request header 0x%08x\n",
		c->id, c->name, h);
	return(-1);
}

// ***************************************************************************
// 
// ***************************************************************************
static void serv_resched_running(server *g, node *c)
{
	job *t, *tnext;
	
	t = c->running.head;
	while (t) {
		tnext = t->next;
		
		// close, remove files, move to pending
		job_close(t, 1);
		queue_unlink(&c->running, t);
		queue_prepend(&g->pending, t);
		
		g->modified = 1;
		
		DEBUG("[%d %s %s]: requeueing\n", t->id, t->dir, t->name);
		
		t = tnext;
	}
}

static void serv_sched_go(server *g, node *c, job *t)
{
	queue_unlink(&g->pending, t);
	queue_append(&c->running, t);
	
	g->modified = 1;
	
	job_open(t);
	
	proto_append(&c->outbuff, PROTO_SV_MAGIC, PROTO_SV_RUN, "isssss",
		t->id, t->shell, t->dir, t->options, t->name, t->cmd);
	
	DEBUG("[%d %s %s]: running on (%d %s)\n",
		t->id, t->dir, t->name, c->id, c->name);
}

// ***************************************************************************
// 
// ***************************************************************************
static int serv_sched_task(server *g, job *t)
{
	node *c, *cbest;
	int rbest;
	
	// Restricted sync task
	if (t->restrict_box >= 0) {
		c = nodeset_find(&g->clients, t->restrict_box);
		
		if (!c) {
			DEBUG("[%d %s %s]: restricted to missing node (%d)\n",
				t->id, t->dir, t->name, t->restrict_box);
			serv_fwd_unref(g, t);
			queue_unlink(&g->pending, t);
			g->modified = 1;
			job_free(t);
			return 0;
		}
		
		if (t->flags & PROTO_JOB_SYNC) {
			c->flags |= NODE_SYNC_PENDING;
			if ((c->running.size == 0) && (c->processes > 0)) {
				c->flags |= NODE_SYNC_RUNNING;
				serv_sched_go(g, c, t);
			}
		} else {
			if ((!(c->flags & NODE_SYNC_ANY))
			&&  (c->running.size < c->processes))
				serv_sched_go(g, c, t);
		}
		return 0;
	}
	
	if (t->flags & PROTO_JOB_SYNC) {
		c = g->clients.head;
		while (c) {
			if (c->running.size > 0)
				return -1;
			c = c->next;
		}
		
		cbest = NULL;
		rbest = 0;
		c = g->clients.head;
		while (c) {
			c->flags |= NODE_SYNC_GLOBAL;
			if ((c->processes > 0) && (c->processes > rbest)) {
				cbest = c;
				rbest = c->processes;
			}
			c = c->next;
		}
		
		if (!cbest)
			return -1;

		cbest->flags |= NODE_SYNC_RUNNING;
		serv_sched_go(g, cbest, t);
		return 0;
	} else {
		cbest = NULL;
		rbest = INT_MAX;
		c = g->clients.head;
		while (c) {
			if ((!(c->flags & NODE_SYNC_ANY))
			&&  (c->running.size < c->processes)
			&&  (c->running.size < rbest)) {
				cbest = c;
				rbest = c->running.size;
			}
			c = c->next;
		}
		
		if (!cbest)
			return -1;
		
		serv_sched_go(g, cbest, t);
		return 0;
	}
}

// ***************************************************************************
// 
// ***************************************************************************
static int server_sched(server *g)
{
	node *c, *cnext;
	job *t, *tnext;
	int runcount;
	
	// per-box handling
	c = g->clients.head;
	while (c) {
		cnext = c->next;
		
		if (serv_sched_client(g, c))
			c->flags |= NODE_KICK;
		
		c->flags &= ~NODE_SYNC_PENDING;
		
		if (c->running.size == 0) {
			if (c->flags & NODE_RETIRE) {
				DEBUG("(%d %s): retiring\n",
					c->id, c->name);
				c->flags |= NODE_KICK;
			}
			
			if ((c->flags & NODE_SYNC_RUNNING)
			&&  (c->flags & NODE_SYNC_GLOBAL)) {
				node *r = g->clients.head;
				
				while (r) {
					r->flags &= ~NODE_SYNC_GLOBAL;
					r = r->next;
				}
			}
			
			c->flags &= ~NODE_SYNC_RUNNING;
		}

		// waste collection, part 1
		if (c->flags & NODE_KICK) {
			DEBUG("(%d %s): disconnect (had %d processes)\n",
				c->id, c->name, c->processes);
			serv_resched_running(g, c);
			close(c->fd);
			nodeset_detach(&g->clients, c);
			node_free(c);
		}
		
		c = cnext;
	}
	
	// per-task handling
	t = g->pending.head;
	while (t) {
		tnext = t->next;

		if (serv_sched_task(g, t))
			break;

		t = tnext;
	}
	
	// waste collection, part 2
	runcount = 0;
	
	c = g->clients.head;
	while (c) {
		cnext = c->next;

		runcount += c->running.size;
		
		if (buffer_io_complete(&c->outbuff)) {
			buffer_reset(&c->outbuff);
			buffer_io_reset(&c->outbuff);
			
			if (c->refs <= 0) {
				DEBUG("(%d %s): done forwarding "
					"(had %d processes)\n",
					c->id, c->name, c->processes);
				// Clients that are referenced should be
				// only capable of reading and displaying
				// data from server. Thus, they have no tasks.
				// So no 'serv_resched_running(g, c);'.
				// Btw, if they had, it would be a problem
				// as we would then need now to reschedule.
				close(c->fd);
				nodeset_detach(&g->clients, c);
				node_free(c);
			}
		}
		
		c = cnext;
	}
	
	if (g->target & SERVER_TARGET_EXIT) {
		DEBUG("Exiting with %d clients, "
			"%d jobs running and %d jobs pending.\n",
			g->clients.n, runcount, g->pending.size);
		return(-1);
	}
	
	if (g->target & SERVER_TARGET_RETIRE) {
		if ((runcount == 0) && (g->pending.size == 0)) {
			DEBUG("Retiring.\n");
			return(-1);
		}
	}

	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
static int server_io(server *g)
{
	struct pollfd *pfd;
	node *c;
	int i, cfd;
	
	// poll
	pfd = (struct pollfd *)array_get(&g->pfd_a, 1 + g->clients.n);
	
	pfd[0].fd = g->fd;
	pfd[0].events = POLLIN;
	pfd[0].revents = 0;
	
	c = g->clients.head;
	for (i = 0; i < g->clients.n; i++) {
		pfd[1 + i].fd = c->fd;
		pfd[1 + i].events = POLLIN;
		if (!buffer_io_complete(&c->outbuff))
			pfd[1 + i].events |= POLLOUT;
		pfd[1 + i].revents = 0;

		c = c->next;
	}
	
	gtime tout = gtime_ms(conf.keepalive_ms);
	
	if ((g->snapshot_path) && (g->modified)) {
		gtime tsnap = gtime_ms(conf.snapshot_ms);
		
		if (gtime_ge(tout, tsnap))
			tout = tsnap;
	}
	
	if (ppoll(pfd, 1 + g->clients.n, &tout, NULL) < 0) {
		fprintf(stderr, "server_io(): poll(): %s\n", strerror(errno));
		return(-1);
	}
	
	g->wall_time_cur = gtime_wall();
	g->mono_time_cur = gtime_mono();
	
	// client i/o
	c = g->clients.head;
	for (i = 0; i < g->clients.n; i++) {
		if (pfd[1 + i].revents & (POLLIN|POLLERR|POLLHUP)) {
			buffer_io_read(&c->inbuff, c->fd);
			c->atime = g->mono_time_cur;
		}
		
		if (pfd[1 + i].revents & (POLLOUT|POLLERR|POLLHUP)) {
			buffer_io_write(&c->outbuff, c->fd);
			c->atime = g->mono_time_cur;
		}

		if ((c->inbuff.eof) || (c->inbuff.error)
		||  (c->outbuff.eof) || (c->outbuff.error)) {
			if (c->inbuff.eof)
				DEBUG("(%d %s): connection closed (in)\n",
					c->id, c->name);
			if (c->outbuff.eof)
				DEBUG("(%d %s): connection closed (out)\n",
					c->id, c->name);
			DEBUG("(%d %s): i/o terminated\n", c->id, c->name);
			c->flags |= NODE_KICK;
		}
		
		c = c->next;
	}
	
	// accept
	if (pfd[0].revents & (POLLIN|POLLERR|POLLHUP)) {
		cfd = socket_accept(g->fd);
		
		if (cfd >= 0) {
			c = node_new(g->box_pool++, cfd);
			nodeset_attach(&g->clients, c);
			c->refs = 1;
			c->atime = g->mono_time_cur;
			DEBUG("(%d %s): connected, fd %d\n",
				c->id, c->name, cfd);
		} else {
			if ((errno == EINTR) || (errno == EAGAIN)) {
				errno = 0;
			} else {
				return(-1);
			}
		}
	}
	
	return(0);
}


// ***************************************************************************
// 
// ***************************************************************************
int ben_server(char *addr, char *snapshot_path, int flags)
{
	server g;
	int r = 0;
	
	g.fd = socket_bind(addr);
	
	if (g.fd < 0)
		return(1);

	if (flags & BEN_FLAG_DAEMONIZE) {
		if (exec_daemonize(NULL, NULL, conf.server_log)) {
			close(g.fd);
			return(1);
		}
	}
	
	nodeset_init(&g.clients);
	queue_init(&g.done);
	queue_init(&g.pending);
	array_init(&g.pfd_a, sizeof(struct pollfd), 0);
	
	g.wall_time_cur = gtime_zero();
	g.mono_time_cur = gtime_zero();
	g.mono_time_snap = gtime_zero();
	
	g.task_pool = 0;
	g.box_pool = 0;
	g.target = SERVER_TARGET_RUNNING;
	g.modified = 0;
	g.snapshot_path = snapshot_path;
	
	while (1) {
		if (server_io(&g)) {
			r = -1;
			break;
		}
		
		int sr = server_sched(&g);
		
		gtime snap_time = gtime_add(g.mono_time_snap,
			gtime_ms(conf.snapshot_ms));
		
		if ((g.snapshot_path) && (g.modified)
		&&  (gtime_ge(g.mono_time_cur, snap_time))) {
			if (server_snapshot(&g.clients, &g.done, &g.pending,
					g.snapshot_path)) {
				r = -1;
				break;
			}
			
			g.modified = 0;
			g.mono_time_snap = g.mono_time_cur;
		}
		
		if (sr) {
			r = 0;
			break;
		}
	}
	
	nodeset_clear(&g.clients);
	queue_clear(&g.done);
	queue_clear(&g.pending);
	array_clear(&g.pfd_a);
	
	socket_unbind(g.fd, addr);
	
	return(r);
}



