/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2024  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include "xmem.h"
#include "buffer.h"
#include "exec.h"
#include "config.h"
#include "units.h"


// ***************************************************************************
// 
// ***************************************************************************
unit *unit_new(int id,
	char *shell, char *dir, char *options, char *name, char *cmd)
{
	unit *u;
	char *av[2];
	char *ae[10];
	
	u = xmalloc(sizeof(unit));
	
	u->id = id;
	u->flags = 0;
	
	av[0] = shell;
	av[1] = NULL;
	
	ae[0] = "outdir";
	ae[1] = dir;
	ae[2] = "options";
	ae[3] = options;
	ae[4] = "task";
	ae[5] = name;
	ae[6] = "job";
	ae[7] = name;
	ae[8] = NULL;
	ae[9] = NULL;
	
	if (exec_fork(&u->exe, av, ae, EXEC_GROUP_LEADER)) {
		free(u);
		return(NULL);
	}

	buffer_init(&u->inbuff, 0);
	buffer_init(&u->outbuff, 0);
	buffer_init(&u->errbuff, 0);
	
	buffer_strcpy(&u->inbuff, cmd);
	buffer_get(&u->outbuff, conf.block_size);
	buffer_get(&u->errbuff, conf.block_size);
	
	return(u);
}

// ***************************************************************************
// 
// ***************************************************************************
void unit_free(unit *u)
{
	exec_terminate(&u->exe, NULL);
	
	buffer_clear(&u->inbuff);
	buffer_clear(&u->outbuff);
	buffer_clear(&u->errbuff);
	
	xfree(u);
}

// ***************************************************************************
// 
// ***************************************************************************
void ul_init(unit_list *l)
{
	l->head = NULL;
	l->tail = NULL;
	l->size = 0;
}

void ul_clear(unit_list *l)
{
	unit *u, *unext;
	
	u = l->head;
	
	while (u) {
		unext = u->next;
		
		unit_free(u);
		
		u = unext;
	}
	
	l->head = NULL;
	l->tail = NULL;
	l->size = 0;
}

// ***************************************************************************
// 
// ***************************************************************************
void ul_prepend(unit_list *l, unit *u)
{
	u->prev = NULL;
	u->next = l->head;
	
	if (l->head)
		l->head->prev = u;
	
	l->head = u;
	
	if (!l->tail)
		l->tail = u;
	
	l->size++;
}

void ul_append(unit_list *l, unit *u)
{
	u->prev = l->tail;
	u->next = NULL;
	
	if (l->tail)
		l->tail->next = u;
	
	l->tail = u;
	
	if (!l->head)
		l->head = u;
	
	l->size++;
}

void ul_unlink(unit_list *l, unit *u)
{
	if (u->prev)
		u->prev->next = u->next;
	
	if (u->next)
		u->next->prev = u->prev;

	if (l->head == u)
		l->head = u->next;
	
	if (l->tail == u)
		l->tail = u->prev;
	
	u->prev = NULL;
	u->next = NULL;
	
	l->size--;
}

// ***************************************************************************
// 
// ***************************************************************************
unit *ul_find(unit_list *l, int id, int flags)
{
	unit *u;
	
	u = l->head;
	
	while (u) {
		if ((u->id == id) && (u->flags == flags))
			return(u);
		
		u = u->next;
	}
	
	return(NULL);
}


